import unittest

import torch

from src.compel.conditioning_scheduler import StaticConditioningScheduler, ConditioningScheduler
from prompting_test_utils import DummyTokenizer, DummyTransformer, KNOWN_WORDS, KNOWN_WORDS_TOKEN_IDS

from src.compel.compel import Compel


def make_dummy_compel():
    tokenizer = DummyTokenizer()
    text_encoder = DummyTransformer()
    return Compel(tokenizer=tokenizer, text_encoder=text_encoder)


def make_test_conditioning(text_encoder: DummyTransformer, tokenizer: DummyTokenizer, token_ids: list[int]) -> torch.Tensor:
    pre_padding = [tokenizer.bos_token_id]
    token_ids = token_ids[0:tokenizer.model_max_length-2]
    post_padding = [tokenizer.eos_token_id] * (tokenizer.model_max_length - len(token_ids) - 1)
    token_ids = pre_padding + token_ids + post_padding
    assert len(token_ids) == tokenizer.model_max_length
    conditioning =  text_encoder(input_ids=torch.tensor(token_ids, dtype=torch.int).unsqueeze(0)).last_hidden_state
    return conditioning


class TestPromptToEmbeddings(unittest.TestCase):

    def test_basic_prompt(self):
        tokenizer = DummyTokenizer()
        text_encoder = DummyTransformer()
        incite = Compel(tokenizer=tokenizer, text_encoder=text_encoder)

        # test "a b c" makes it to the Conditioning intact for t=0, t=0.5, t=1
        prompt = " ".join(KNOWN_WORDS[:3])
        conditioning_scheduler = incite.make_conditioning_scheduler(prompt)
        conditioning_scheduler_2 = incite.make_conditioning_scheduler(prompt)
        expected_positive_conditioning = make_test_conditioning(text_encoder, tokenizer, KNOWN_WORDS_TOKEN_IDS[:3])
        expected_negative_conditioning = make_test_conditioning(text_encoder, tokenizer, [])
        self.assert_constant_scheduling_matches_expected(conditioning_scheduler,
                                                         expected_positive_conditioning,
                                                         expected_negative_conditioning)


    def test_basic_negative_prompt(self):
        tokenizer = DummyTokenizer()
        text_encoder = DummyTransformer()
        incite = Compel(tokenizer=tokenizer, text_encoder=text_encoder)

        # positive "a b c" negative "c b a" makes it to the Conditioning intact for t=0, t=0.5, t=1
        positive_prompt = " ".join(KNOWN_WORDS[:3])
        negative_prompt = " ".join(reversed(KNOWN_WORDS[:3]))
        conditioning_scheduler = incite.make_conditioning_scheduler(positive_prompt, negative_prompt)
        expected_positive_conditioning = make_test_conditioning(text_encoder, tokenizer, KNOWN_WORDS_TOKEN_IDS[:3])
        expected_negative_conditioning = make_test_conditioning(text_encoder, tokenizer, list(reversed(KNOWN_WORDS_TOKEN_IDS[:3]))
        )
        self.assert_constant_scheduling_matches_expected(conditioning_scheduler,
                                                         expected_positive_conditioning,
                                                         expected_negative_conditioning)

    def test_too_long_prompt(self):
        tokenizer = DummyTokenizer()
        text_encoder = DummyTransformer()
        incite = Compel(tokenizer=tokenizer, text_encoder=text_encoder)

        # positive "a b c" negative "c b a" makes it to the Conditioning intact for t=0, t=0.5, t=1
        positive_prompt = " ".join(KNOWN_WORDS[:3] * 40)
        conditioning_scheduler = incite.make_conditioning_scheduler(positive_prompt)
        expected_positive_conditioning = make_test_conditioning(text_encoder, tokenizer, KNOWN_WORDS_TOKEN_IDS[:3] * 40)
        expected_negative_conditioning = make_test_conditioning(text_encoder, tokenizer, [])
        self.assert_constant_scheduling_matches_expected(conditioning_scheduler,
                                                         expected_positive_conditioning,
                                                         expected_negative_conditioning)


    def assert_constant_scheduling_matches_expected(self,
                                                    conditioning_scheduler: ConditioningScheduler,
                                                    expected_positive_conditioning: torch.Tensor,
                                                    expected_negative_conditioning: torch.Tensor):
        self.assertIs(StaticConditioningScheduler, type(conditioning_scheduler))

        conditioning_at_start = conditioning_scheduler.get_conditioning_for_step_pct(0)
        self.assertTrue(torch.allclose(expected_positive_conditioning,
                                       conditioning_at_start.positive_conditioning,
                                       atol=1e-6))
        self.assertTrue(torch.allclose(expected_negative_conditioning,
                                       conditioning_at_start.negative_conditioning,
                                       atol=1e-6))

        conditioning_at_mid = conditioning_scheduler.get_conditioning_for_step_pct(0.5)
        self.assertTrue(torch.allclose(expected_positive_conditioning,
                                       conditioning_at_mid.positive_conditioning,
                                       atol=1e-6))
        self.assertTrue(torch.allclose(expected_negative_conditioning,
                                       conditioning_at_mid.negative_conditioning,
                                       atol=1e-6))

        conditioning_at_end = conditioning_scheduler.get_conditioning_for_step_pct(1.0)
        self.assertTrue(torch.allclose(expected_positive_conditioning,
                                       conditioning_at_end.positive_conditioning,
                                       atol=1e-6))
        self.assertTrue(torch.allclose(expected_negative_conditioning,
                                       conditioning_at_end.negative_conditioning,
                                       atol=1e-6))




if __name__ == '__main__':
    unittest.main()



from typing import Union

import torch
from transformers import CLIPTokenizer, CLIPTextModel

from .conditioning_scheduler import ConditioningScheduler, StaticConditioningScheduler
from .embeddings_provider import EmbeddingsProvider
from .prompt_parser import Blend, FlattenedPrompt, PromptParser

__all__ = ["Compel"]

class Compel:

    def __init__(self, tokenizer: CLIPTokenizer, text_encoder: CLIPTextModel):
        self.conditioning_provider = EmbeddingsProvider(tokenizer=tokenizer, text_encoder=text_encoder)

    def make_conditioning_scheduler(self, positive_prompt: str, negative_prompt: str='') -> ConditioningScheduler:
        positive_conditioning = self.build_conditioning_tensor(positive_prompt)
        negative_conditioning = self.build_conditioning_tensor(negative_prompt)
        return StaticConditioningScheduler(positive_conditioning=positive_conditioning,
                                           negative_conditioning=negative_conditioning)

    def build_conditioning_tensor(self, text: str) -> torch.Tensor:
        prompt_object = self.parse_prompt_string(text)
        conditioning = self.build_conditioning_tensor_for_prompt_object(prompt_object)
        return conditioning

    @classmethod
    def parse_prompt_string(cls, prompt_string: str) -> Union[FlattenedPrompt, Blend]:
        pp = PromptParser()
        conjunction = pp.parse_conjunction(prompt_string)
        # we don't support conjunctions for now
        parsed_prompt = conjunction.prompts[0]
        return parsed_prompt

    def describe_tokenization(self, text: str) -> list[str]:
        """
        For the given text, return a list of strings showing how it will be tokenized.

        :param text: The text that is to be tokenized.
        :return: A list of strings representing the output of the tokenizer. It's expected that the output list may be
        longer than the number of words in `text` because the tokenizer may split words to multiple tokens. Because of
        this, word boundaries are indicated in the output with `</w>` strings.
        """
        return self.conditioning_provider.tokenizer.tokenize(text)

    def build_conditioning_tensor_for_prompt_object(self, prompt: Union[Blend, FlattenedPrompt]) -> torch.Tensor:
        """

        """
        if type(prompt) is Blend:
            return self._get_conditioning_for_blend(prompt)
        elif type(prompt) is FlattenedPrompt:
            if prompt.wants_cross_attention_control:
                raise NotImplementedError("Not implemented yet - see InvokeAI for reference implementation")
                # conditioning, cac_args = self._get_conditioning_for_cross_attention_control(model, parsed_prompt, log_tokens)
            else:
                return self._get_conditioning_for_flattened_prompt(prompt)

        raise ValueError(f"unsupported prompt type: {type(prompt).__name__}")

    def _get_conditioning_for_flattened_prompt(self, prompt: FlattenedPrompt) -> torch.Tensor:
        if type(prompt) is not FlattenedPrompt:
            raise ValueError(f"embeddings can only be made from FlattenedPrompts, got {type(prompt).__name__} instead")
        fragments = [x.text for x in prompt.children]
        weights = [x.weight for x in prompt.children]
        conditioning = self.conditioning_provider.get_embeddings_for_weighted_prompt_fragments(text_batch=[fragments],
                                                                                               fragment_weights_batch=[
                                                                                                   weights])
        return conditioning

    def _get_conditioning_for_blend(self, blend: Blend):
        conditionings_to_blend = None
        for i, flattened_prompt in enumerate(blend.prompts):
            this_conditioning = self._get_conditioning_for_flattened_prompt(flattened_prompt)
            conditionings_to_blend = this_conditioning if conditionings_to_blend is None else torch.cat(
                (conditionings_to_blend, this_conditioning))
        conditioning = EmbeddingsProvider.apply_embedding_weights(conditionings_to_blend.unsqueeze(0),
                                                                  blend.weights,
                                                                  normalize=blend.normalize_weights)
        return conditioning





from dataclasses import dataclass

import torch

__all__ = ["ConditioningScheduler", "Conditioning", "StaticConditioningScheduler"]

@dataclass
class Conditioning():
    """
    Conditioning. In all examples `B` is batch size, `77` is the text encoder's max token length,
    and `token_dim` is 768 for SD1 and 1280 for SD2.
    """
    positive_conditioning: torch.Tensor # shape [B x 77 x token_dim]
    negative_conditioning: torch.Tensor # shape [B x 77 x token_dim]


class ConditioningScheduler():
    """
    Provides a mechanism to control which processes to apply for any given step of a Stable Diffusion generation.
    """

    def get_conditioning_for_step_pct(self, step_pct: float) -> Conditioning:
        """
        Return the conditioning to apply at the given step.
        :param step_pct: The step as a float `0..1`, where `0.0` is immediately before the start of image generation
        process (when the latent vector is 100% noise), and `1.0` is immediately after the end of the final step
        (when the latent vector represents the final noise-free generated image).
        :return: The Conditioning to apply for the requested step.
        """
        raise NotImplementedError("Subclasses must override")


class StaticConditioningScheduler(ConditioningScheduler):
    def __init__(self, positive_conditioning: torch.Tensor,
                 negative_conditioning: torch.Tensor):
        self.positive_conditioning = positive_conditioning
        self.negative_conditioning = negative_conditioning

    def get_conditioning_for_step_pct(self, step_pct: float) -> Conditioning:
        """ See base class for docs. """
        return Conditioning(positive_conditioning=self.positive_conditioning,
                            negative_conditioning=self.negative_conditioning,
                            )
